/***************************************************************************/
/* MsgPost Version 1.00 Jan 1992           (C) 1992 by CodeLand Australia, */
/*                                                    All Rights Reserved. */
/* Written by Colin Wheat of Fidonet 3:690/613                             */
/* Compiled using Turbo C++ v1.00                                          */
/*                                                                         */
/* MsgPost uses the Squish Message Base Level 0 MsgAPI                     */
/* Squish is a trademark of Scott J. Dudley                                */
/*                                                                         */
/* Command line:                                                           */
/* -T<name>  Message text file path and name                               */
/* -C<name>  Configuration file path and name                              */
/* -M<name>  Message area path and name                                    */
/* -N<addr>  Netmail format - Send to address\n"                           */
/* -S[##]    Split long messages to ## Kb size (0 - 16)                    */
/* -PX[X]    Message priority flag(s)                                      */
/* -K        Kill text file after processing                               */
/* -?        Program help screen                                           */
/*                                                                         */
/* Error levels:                                                           */
/* 0 - Normal exit                                                         */
/* 1 - Syntax exit                                                         */
/* 2 - Out of memory                                                       */
/* 3 - Text file not found                                                 */
/* 4 - No system address set                                               */
/* 5 - Text file is blank                                                  */
/* 6 - Message base open failed                                            */
/*                                                                         */
/***************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <io.h>
#include <dir.h>
#include <dos.h>
#include "prog.h"               /* Squish API header */
#include "alc.h"                /* Squish API header */
#include "msgapi.h"             /* Squish API header */

#define VERSION   "v1.00"       /* MsgPost version   */
#define MAX_BLOCK 16000         /* Maximum body size */
#define MAX_LINE  10000         /* Maximum lines     */

int kill_txt=0;                 /* Txtfile kill flag */
int netmail=0;                  /* Netmail area      */
int ga=0;                       /* Got address flag  */
int split_k=12;                 /* Message split     */
unsigned long seed;             /* MSGID seed        */
time_t time_now;                /* Creation time     */
int mn;                         /* Split unique #    */

dword attr=MSGLOCAL;            /* Message attr      */
struct _netaddr myaddr;         /* System address    */
struct _netaddr toaddr;         /* Netmail to addr   */

char homepath[80];              /* Executable path   */
char msgarea[80];               /* Message area name */
char msgtext[80];               /* Import text file  */
char cfgname[80];               /* MsgPost CFG file  */

char str_to[XMSG_TO_SIZE];
char str_from[XMSG_FROM_SIZE];
char str_subj[XMSG_SUBJ_SIZE];
char str_orig[80];
char *def_orig = "^KD <Esc> <F2> ^Z ^C (damnit!) q e quit exit !Q system";

int linesnum=0;
char *lines[MAX_LINE];
long linesbytes=0L;

long body_len;                  /* Body length       */
char *body;                     /* Message body      */

static void pascal write_msg (MSG *ap);
static int pascal get_num (int splitbytes);
static int pascal build_body (int limit);
static void pascal read_cfg (void);
static int pascal read_txt (void);
static int pascal set_msgcfg (char *s);
static void pascal set_attr (char *p);
static int pascal read_orig (void);
static void pascal build_tear (char *s);
static void pascal build_hdr (XMSG *x, int num, int maxnum);
static void pascal build_ctrl (char *str, int *len, int num, int maxnum);
static unsigned long pascal hsec_time (void);
static void pascal get_addr (char *str, struct _netaddr *addr);
static void pascal add_slash (char *str);
static void pascal strip_slash(char *str);
static void pascal strip_cr (char *str);
static void pascal strtrim (char *str);
static int pascal strblank (char *str);
static void pascal cvt_us (char *s);
static int pascal is_space (char c);
static char *pascal fancy_s (char *string);
static void pascal make_ourpath (char *pth);
static void pascal setup (int argc, char *argv[]);
static void pascal parse_cmd_line (int argc, char *argv[]);
static void pascal syntax_exit (void);


void main (int argc, char *argv[])
{
    int areatyp;                /* Message area type */
    struct _minf mi;            /* API structure     */
    MSG *ap;                    /* API area pointer  */

    printf("\nMsgPost " VERSION " - the Fidonet/Squish Message Base Writer"
           "\n   (C) Copyright 1992 by CodeLand, All Rights Reserved\n\n");

    setup(argc,argv); read_cfg(); /* Setup */
    if(!read_txt()) { printf("\nError; Text file not found!\n\n"); exit(3); }
    if(!linesbytes) { printf("\nError; Text file is blank!\n\n"); exit(5); }
    parse_cmd_line(argc,argv); /* Get command line overrides */

    /* Check we have a system address */
    if(!ga) { printf("\nError; No address set!\n\n"); exit(4); }
    mi.def_zone=myaddr.zone; mi.req_version=0; /* Set api */

    /* If no origin set, try from the message area, else use default */
    if(!*str_orig) if(!read_orig()) strcpy(str_orig,def_orig);

    MsgOpenApi(&mi);
    areatyp=*msgarea=='$'?MSGTYPE_SQUISH:MSGTYPE_SDM;
    if(areatyp==MSGTYPE_SDM&&!netmail) areatyp|=MSGTYPE_ECHO;

    /* Open the message base */
    if((ap=MsgOpenArea((byte *)msgarea+(*msgarea=='$'),MSGAREA_NORMAL,areatyp)
       )==NULL) {
        printf("\nError; Message base open failed!\n\n");
        exit(6);
    }

    MsgLock(ap);            /* Lock the base        */
    write_msg(ap);          /* Write the message(s) */
    MsgCloseArea(ap);       /* Close the base       */

    MsgCloseApi();          
    if(kill_txt) unlink(msgtext);
    printf("\nDone!\n");
    exit(0);
}


/*
** write_msg ()
** write message(s)
*/

static void pascal write_msg (MSG *ap)
{
    register i;
    register split_num=1;       /* How many msgs     */
    register splitbytes;        /* Split size        */
    MSGH *msg;                  /* Message pointer   */
    XMSG xmsg;                  /* Xmsg structure    */
    int ctrllen;                /* Control info size */
    char ctrl[256];             /* Control info      */
    char tear[256];             /* Tear line, origin */

    time_now=time(NULL); /* Get message time */
    seed=hsec_time();    /* Get MSGID seed */
    srand((unsigned)seed); mn=rand(); /* Get split messages unique number */
    build_tear(tear); /* Make tear/origin line for echomail */

    /* Set each message size */
    if(!split_k) splitbytes=MAX_BLOCK;
    else splitbytes=split_k*1024;
    if(splitbytes>MAX_BLOCK) splitbytes=MAX_BLOCK;

    /* Adjust for tear and header */
    if(!netmail) splitbytes-=strlen(tear);
    splitbytes-=250;

    if(linesbytes>(long)splitbytes) { /* Split or cut */
        if(split_k) {
            split_num=get_num(splitbytes); /* Compute number of messages */
            /* Set to average message size for neatness :-) */
            splitbytes=(int)(linesbytes/(long)split_num);
            for(;;) { /* And adjust size back up so no text is lost */
                if(get_num(splitbytes)<=split_num) break;
                splitbytes++;
            }
        }
    }

    printf("Writing \"%s\"  01/%02u ",fancy_s(msgarea),split_num);
    for(i=0;i<split_num;i++) {
        printf("\b\b\b\b\b\b%02u/%02u ",i+1,split_num);
        if(!build_body(splitbytes)) break;
        if(!netmail) { strcat(body,tear); body_len+=(long)strlen(tear); }
        if(i) attr&=~MSGFILE; /* Kill file attaches on 2nd, 3rd, etc */
        build_hdr(&xmsg,i+1,split_num);
        msg=MsgOpenMsg(ap,MOPEN_CREATE,0L); /* Create new message */
        build_ctrl(ctrl,&ctrllen,i+1,split_num);
        MsgWriteMsg(msg,FALSE,&xmsg,NULL,0L,body_len,ctrllen,(byte *)ctrl);
        MsgWriteMsg(msg,TRUE,NULL,(byte *)body,body_len,body_len,0L,NULL);
        MsgCloseMsg(msg);
    }
    printf("\n");
}


/*
** get_num ()
** Compute how many split messages to generate
*/

static int pascal get_num (int splitbytes)
{
    register i, result=1;
    long len=0L;

    /* Scan through the lines */
    for(i=0;i<linesnum;i++) {
        len+=(long)strlen(lines[i]);
        if(len>(long)splitbytes) {
            result++;
            len=(long)strlen(lines[i]);
        }
    }
    return result;
}


/*
** build_body ()
** Fill the body buffer from the lines array
*/

static int pascal build_body (int limit)
{
    register len;
    static int linescount=0;
    char *p;

    if(linescount>=linesnum) return 0;
    p=body; body_len=0L;

    for(;;) {
        len=strlen(lines[linescount]);
        if(body_len+(long)len>limit) break;
        memcpy(p,lines[linescount],len);
        p+=len; body_len+=(long)len;
        if(++linescount>=linesnum) break;
    }
    *p='\0'; body_len++;

    return 1;
}


/*
** read_cfg ()
** read configuration file
*/

static void pascal read_cfg (void)
{
    FILE *fp;
    char line[258];

    printf("Reading \"%s\" ",fancy_s(cfgname));
    if((fp=fopen(cfgname,"r"))==NULL) {
        printf("(Not found)\n");
        return;
    }
    while(fgets(line,256,fp)) if(*line!=';') set_msgcfg(line);
    fclose(fp);
    printf("\n");
}


/*
** read_txt ()
** read text file into memory
*/

static int pascal read_txt (void)
{
    register llen, checkcfg=1, trimleading=1;
    FILE *fp;
    char *fbuf=NULL;
    char line[258];

    printf("Reading \"%s\"\n",fancy_s(msgtext));
    if((fp=fopen(msgtext,"r"))==NULL) return 0;
    fbuf=(char *)malloc(4096); /* Get extended file buffer */
    if(fbuf!=NULL) setvbuf(fp,fbuf,_IOFBF,4096);

    while(fgets(line,256,fp)) {

        if(checkcfg) { /* Check for override configuration */
            if(*line==';') continue;
            if(set_msgcfg(line)) continue;
            else checkcfg=0;
        }

        strip_cr(line); strtrim(line); /* Clean up */

        /* Ignore leading blank lines */
        if(trimleading) {
            if(strblank(line)) continue;
            else trimleading=0;
        }

        strcat(line,"\r"); llen=strlen(line);
        if((lines[linesnum]=(char *)malloc(llen+1))==NULL) {
            printf("\nError; Out of memory!\n\n");
            fclose(fp);
            exit(2);
        }

        memcpy(lines[linesnum],line,llen+1);
        linesbytes+=(long)llen; /* Add to total body bytes */
        if(++linesnum>=MAX_LINE) break;
    }

    /* Strip trailing blank lines */
    for(;;) { /* Step back */
        if(!strblank(lines[linesnum-1])) break;
        linesbytes-=(long)strlen(lines[linesnum-1]);
        if(!--linesnum) break;
    }

    fclose(fp);
    if(fbuf) free(fbuf);
    return 1;
}


/*
** set_msgcfg ()
** Set message configuration from configuration string
*/

static int pascal set_msgcfg (char *s)
{
    register status=0;
    char *p, *q, *r, sbuf[258];

    strcpy(sbuf,s); /* Clone the line so we don't change the original */

    if((p=strtok(sbuf," \t\n\r"))!=NULL) {
        if((q=strtok(NULL,"\n\r"))!=NULL) {
            r=q; while(*r==' '||*r=='\t') r++; /* Strip leading */
            strtrim(r); /* Strip trailing */
            if(*r) {
                if(!stricmp(p,"To:")) {
                    strncpy(str_to,r,XMSG_TO_SIZE-1);
                    str_to[XMSG_TO_SIZE-1]='\0';
                    status=1;
                }
                else if(!stricmp(p,"From:")) {
                    strncpy(str_from,r,XMSG_FROM_SIZE-1);
                    str_from[XMSG_FROM_SIZE-1]='\0';
                    status=1;
                }
                else if(!stricmp(p,"Subj:")) {
                    strncpy(str_subj,r,XMSG_SUBJ_SIZE-1);
                    str_subj[XMSG_SUBJ_SIZE-1]='\0';
                    status=1;
                }
                else if(!stricmp(p,"Origin:")) {
                    strncpy(str_orig,r,59);
                    str_orig[59]='\0';
                    status=1;
                }
                else if(!stricmp(p,"Address:")) {
                    get_addr(r,&myaddr); ga++;
                    status=1;
                }
                else if(!stricmp(p,"Attr:")) {
                    set_attr(r);
                    status=1;
                }
                else if(!stricmp(p,"Area:")) {
                    strncpy(msgarea,r,78); msgarea[78]='\0';
                    strip_slash(msgarea);
                    status=1;
                }
                else if(!stricmp(p,"Netmail:")) {
                    get_addr(r,&toaddr); netmail=1;
                    status=1;
                }
            }
        }
    }

    return status;
}


/*
** set_msgattr ()
** Set message attribute flags
*/

static void pascal set_attr (char *p)
{
    while(*p) {
        *p=tolower(*p);
        switch (*p) {
        case 'c': /* Crash */
                  attr|=MSGCRASH;
                  attr&=~MSGHOLD; break;
        case 'f': /* File attach */
                  attr|=MSGFILE; break;
        case 'h': /* Hold */
                  attr|=MSGHOLD;
                  attr&=~MSGCRASH; break;
        case 'k': /* Kill after sending */
                  attr|=MSGKILL; break;
        case 'p': /* Private */
                  attr|=MSGPRIVATE; break;
        case 'r': /* File Request */
                  attr|=MSGFRQ; break;
        case 'u': /* File Update Request */
                  attr|=MSGURQ; break;
        }
        p++;
    }
}


/*
** read_orig ()
** read message area origin file
*/

static int pascal read_orig (void)
{
    FILE *fp;
    char *p, buf[128];

    if(*msgarea=='$') { strcpy(buf,msgarea+1); strcat(buf,".SQO"); }
    else { strcpy(buf,msgarea); strcat(buf,"\\ORIGIN."); }
    printf("Reading \"%s\"\n",fancy_s(buf));

    if((fp=fopen(buf,"r"))==NULL) return 0;
    fgets(buf,128,fp);
    fclose(fp);

    p=buf; while(*p==' '||*p=='\t') p++; /* Strip leading */
    if(strlen(p)>59) p[59]='\0';
    strip_cr(p); strtrim(p);
    strcpy(str_orig,p);
    return 1;
}


/*
** build_tear ()
** read text file
*/

static void pascal build_tear (char *s)
{
    char line[90], sbuf[20];

    *s='\0';
    if(!netmail) {
        /* Tear line */
        strcpy(s,"\r--- MsgPost " VERSION "\r");  
        /* Origin line */
        if(myaddr.point) sprintf(sbuf,".%u",myaddr.point);
        else *sbuf='\0';
        sprintf(line," * Origin: %s (%u:%u/%u%s)\r",
            str_orig,myaddr.zone,myaddr.net,myaddr.node,sbuf);
        strcat(s,line);
        /* Seen-by line */
        sprintf(line,"SEEN-BY: %u/%u\r",myaddr.net,myaddr.node);
        strcat(s,line);
    }
}


/*
** build_hdr ()
** Set message header information
*/

static void pascal build_hdr (XMSG *x, int num, int maxnum)
{
    union stamp_combo combo;
    struct tm *tmdate;

    memset(x,'\0',sizeof(XMSG));

    x->attr=attr; /* Set message attributes */

    x->orig=x->dest=myaddr;
    if(netmail) x->dest=toaddr;

    strcpy((char *)x->to,str_to);
    strcpy((char *)x->from,str_from);

    if(maxnum>1) {
        sprintf((char *)x->subj,"[%d of %d] ",num,maxnum);
        strncat((char *)x->subj,str_subj,
                XMSG_SUBJ_SIZE-1-strlen((char *)x->subj));
        (x->subj)[XMSG_SUBJ_SIZE-1]='\0';
    }
    else strcpy((char *)x->subj,str_subj);

    tmdate=localtime(&time_now);
    x->date_written=x->date_arrived=(TmDate_to_DosDate(tmdate,&combo))->msg_st;
}


/*
** build_ctrl ()
** set message control information
*/

static void pascal build_ctrl (char *str, int *len, int num, int maxnum)
{
    struct tm *t;
    unsigned long time_new;
    char ubuf[20], sbuf[80];
    char *const Months[12] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
                               "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

    *len=0; 
    time_new=hsec_time(); if(seed<time_new) seed=time_new;

    sprintf(str,"\01MSGID: %u:%u/%u.%u %08lx",
        myaddr.zone,myaddr.net,myaddr.node,myaddr.point,seed);

    /* ^ASPLIT: 30 Mar 90 11:12:34 @494/4       12345 02/03 +++++++++++ */
    if(maxnum>1) {
        t=localtime(&time_now);
        sprintf(ubuf,"@%u/%u",myaddr.net,myaddr.node); ubuf[12]='\0';
        sprintf(sbuf,"\01SPLIT: %02u %s %02u %02u:%02u:%02u %-12s %05u %02u/%02u +++++++++++",
            t->tm_mday,Months[t->tm_mon],t->tm_year,
            t->tm_hour,t->tm_min,t->tm_sec,ubuf,mn,num,maxnum);
        strcat(str,sbuf);
    }

    *len=strlen(str)+1; seed++;
}


/*
** hsec_time ()
** get system date/time in hsec's
*/

static unsigned long pascal hsec_time (void)
{
    unsigned long i;
    struct time t;
    struct date d;

    getdate(&d); gettime(&t);
    i=((d.da_day+(((unsigned)d.da_mon*3057-3007)/100))*144000L) +
    (t.ti_hour*360000L) +                             /* hours today: hsec    */
    (t.ti_min*6000L) +                                /* minutes: hsec        */
    (t.ti_sec*100L) +                                 /* seconds: hsec        */
    t.ti_hund;                                        /* hundreds of seconds  */

    return i;
}


/*
** get_addr ()
** get command line system address
*/

static void pascal get_addr (char *str, struct _netaddr *addr)
{
    char *p;

    if((p=strchr(str,':'))!=NULL) {
        addr->zone=atoi(str);
        str=p+1;
    }
    else addr->zone=myaddr.zone;

    if((p=strchr(str,'/'))!=NULL) {
        addr->net=atoi(str);
        str=p+1;
    }
    else addr->net=0;

    if((p=strchr(str,'.'))!=NULL) {
        addr->node=atoi(str);
        str=p+1;
        addr->point=atoi(str);
    }
    else {
        addr->node=atoi(str);
        addr->point=0;
    }
}


/*
** add_slash (char *str)
** Add trailing back slash
*/

static void pascal add_slash (char *str)
{
    char *p;

    p=str; while(*p) p++;
    if(p!=str) { p--; if(*p!='\\') strcat(str,"\\"); }
}


/*
** strip_slash ()
** strip trailing back slash
*/

static void pascal strip_slash(char *str)
{
    register i;

    for(i=strlen(str)-1;(str[i]=='\\')&&i>=0;i--) ;
    str[i+1]='\0';
}


/*
** strip_cr ()
** strip trailing cr/lf
*/

static void pascal strip_cr (char *str)
{
    register i;

    for(i=strlen(str)-1;(str[i]=='\r'||str[i]=='\n')&&i>=0;i--) ;
    str[i+1]='\0';
}


/*
** strtrim ()
** strip trailing white space
*/

static void pascal strtrim (char *str)
{
    register int i;

    for(i=strlen(str)-1;is_space(str[i])&&i>=0;i--) ;
    str[i+1]='\0';
}


/*
** strblank ()
** is string blank?
*/

static int pascal strblank (char *str)
{
    register char *p;
    register int blank=1;

    for(p=str;*p;p++) {
        if(!is_space(*p)) {
            blank=0;
            break;
        }
    }
    return blank;
}


/*
** cvt_us ()
** convert underscore characters to spaces
*/

static void pascal cvt_us (char *s)
{
    while(*s) { if(*s=='_') *s=' '; s++; }
}


/*
** is_space ()
** if the character is whitespace
*/

static int pascal is_space (char c)
{
    if(c==' '||c=='\t') return 1;
    return 0;
}


/*
** fancy_s ()
** Make It Look Like This
*/

static char *pascal fancy_s (char *string)
{
    register flag=0;
    char *s;

    s=string;
    while(*string) {
        if(isalpha(*string)) {                  /* If alphabetic,     */
            if(flag) *string=tolower(*string);  /* already saw one?   */
            else {
                flag = 1;                       /* first one, flag it */
                *string=toupper(*string);       /* Uppercase it       */
            }
        }
        else flag=0;                            /* reset alpha flag   */
        string++;
    }
    return s;
}


/*
** make_ourpath ()
** get executable path
*/

static void pascal make_ourpath (char *pth)
{
    char drive[MAXDRIVE], dir[MAXDIR], name[MAXFILE], ext[MAXEXT];

    if(_osmajor>=3) { /* If DOS 3 or later, use argv[0] */
        fnsplit(pth,drive,dir,name,ext);
        strcpy(homepath,drive); strcat(homepath,dir);
        add_slash(homepath);
    }
    else homepath[0]='\0';
}


/*
** setup ()
** program setup
*/

static void pascal setup (int argc, char *argv[])
{
    register i;
    char *p;

    /* Get the text storage */
    if((body=malloc(MAX_BLOCK+1024))==NULL) {
        printf("\nError; Out of memory!\n\n");
        exit(2);
    }
    memset(body,'\0',MAX_BLOCK+1024); /* Clear it */

    *str_orig=*msgarea=*msgtext='\0';
    make_ourpath(argv[0]);
    strcpy(cfgname,homepath); strcat(cfgname,"MsgPost.Cfg");
    strcpy(str_to,"All");
    strcpy(str_from,"MsgPost " VERSION);
    strcpy(str_subj,"Automated Posting");

    /* Get file names */
    for(i=1;i<argc;i++) {
        p=argv[i];
        if(*p=='-'||*p=='/') {
            switch(tolower(*(++p))) {
                case 'c': strcpy(cfgname,++p); break;
                case 't': strcpy(msgtext,++p); break;
                case 'm':
                case 'n':
                case 'p':
                case 'j':
                case 'k':
                case 's': break;
                case '?': 
                default : syntax_exit();
            }
        }
        else syntax_exit();
    }
}


/*
** parse_cmd_line ()
** get command line parameters
*/

static void pascal parse_cmd_line (int argc, char *argv[])
{
    register i;
    char *p, s[60];

    for(i=1;i<argc;i++) {
        p=argv[i];
        if(*p=='-'||*p=='/') {
            switch(tolower(*(++p))) {
                case 'm': strcpy(msgarea,++p); strip_slash(msgarea); break;
                case 'n': strcpy(s,++p); get_addr(s,&toaddr); netmail=1; break;
                case 'p': set_attr(++p); break;
                case 'j': strncpy(str_subj,++p,XMSG_SUBJ_SIZE-1);
                          str_subj[XMSG_SUBJ_SIZE-1]='\0';
                          cvt_us(str_subj); break;
                case 's': split_k=atoi(++p);
                          if(split_k<0||split_k>16) split_k=12;
                          break;
                case 'k': kill_txt=1; break;
            }
        }
    }
}


/*
** syntax_exit ()
** syntax command line display and exit
*/

static void pascal syntax_exit (void)
{
   puts("    Syntax:  MsgPost [-switch -switch ... ]\n\n"
        "\t -T<name>    Text file path and name\n"
        "\t -C<name>    Optional configuration file path and name\n"
        "\t -M<name>    Message area path and name override\n"
        "\t -J<subj>    Message subject override\n"
        "\t -N<addr>    Netmail format - Send to address\n"
        "\t -S[##]      Split long messages to ## Kb size (0 - 16)\n"
        "\t -P[cfhkpru] Message priority flag(s)\n"
        "\t -K          Kill text file after processing\n"
        "\t -?          Program help screen\n"
    );

    exit(1);
}

